/*
 * EbModas.c
 *
 *  Created on: 13.01.2014
 *      Author: thomas
 */

// http://wiki.openwrt.org/doc/howto/netfilter/netfilter.ebtables.example1 (ebtables ist eine layer2 Technik)
// http://ebtables.sourceforge.net/examples/example5.html
// http://l7-filter.sourceforge.net/L7-Netfilter-example

#include "EbModas.h"

#include <linux/ip.h>
#include <net/ip.h>
#include <linux/in.h>
#include <linux/module.h>
#include <linux/netfilter/x_tables.h>
#include <linux/netfilter_bridge/ebtables.h>

#include <linux/types.h>

#define EBT_IP_SOURCE   0x01
#define EBT_IP_DEST     0x02
#define EBT_IP_TOS      0x04
#define EBT_IP_PROTO    0x08
#define EBT_IP_SPORT    0x10
#define EBT_IP_DPORT    0x20
#define EBT_IP_MASK     (EBT_IP_SOURCE | EBT_IP_DEST | EBT_IP_TOS | EBT_IP_PROTO | EBT_IP_SPORT | EBT_IP_DPORT )
#define EBT_IP_MATCH    "modas"

/* the same values are used for the invflags */
struct EbtIpInfo_s
{
  __be32 saddr;
  __be32 daddr;
  __be32 smsk;
  __be32 dmsk;
  __u8  tos;
  __u8  protocol;
  __u8  bitmask;
  __u8  invflags;
  __u16 sport[2];
  __u16 dport[2];
};

struct EbtTcpUdpHdr_s
{
  __be16 src;
  __be16 dst;
};

static bool EbModasMatchCB(const struct sk_buff *skb, struct xt_action_param *par)
{
  const struct EbtIpInfo_s *info = par->matchinfo;
  const struct iphdr *ih;
  struct iphdr _iph;
  const struct EbtTcpUdpHdr_s *pptr;
  struct EbtTcpUdpHdr_s _ports;

  printk(KERN_INFO "EbModasMatchCheckCB\n");

  ih = skb_header_pointer(skb, 0, sizeof(_iph), &_iph);
  if (ih == NULL)
  {
    return false;
  }
  if (info->bitmask & EBT_IP_TOS && FWINV(info->tos != ih->tos, EBT_IP_TOS))
  {
    return false;
  }
  if (info->bitmask & EBT_IP_SOURCE && FWINV((ih->saddr & info->smsk) != info->saddr, EBT_IP_SOURCE))
  {
    return false;
  }
  if ((info->bitmask & EBT_IP_DEST) && FWINV((ih->daddr & info->dmsk) != info->daddr, EBT_IP_DEST))
  {
    return false;
  }
  if (info->bitmask & EBT_IP_PROTO)
  {
    if (FWINV(info->protocol != ih->protocol, EBT_IP_PROTO))
    {
      return false;
    }
    if (!(info->bitmask & EBT_IP_DPORT) && !(info->bitmask & EBT_IP_SPORT))
    {
      return true;
    }
    if (ntohs(ih->frag_off) & IP_OFFSET)
    {
      return false;
    }
    pptr = skb_header_pointer(skb, ih->ihl*4, sizeof(_ports), &_ports);
    if (pptr == NULL)
    {
      return false;
    }
    if (info->bitmask & EBT_IP_DPORT)
    {
      u32 dst = ntohs(pptr->dst);
      if (FWINV(dst < info->dport[0] || dst > info->dport[1], EBT_IP_DPORT))
      {
        return false;
      }
    }
    if (info->bitmask & EBT_IP_SPORT)
    {
      u32 src = ntohs(pptr->src);
      if (FWINV(src < info->sport[0] || src > info->sport[1], EBT_IP_SPORT))
      {
        return false;
      }
    }
  }
  return true;
}

static int EbModasMatchCheckCB(const struct xt_mtchk_param *par)
{
  const struct EbtIpInfo_s *info = par->matchinfo;
  const struct ebt_entry *e = par->entryinfo;

  printk(KERN_INFO "EbModasMatchCheckCB %04X\n", e->ethproto);

  if (e->ethproto != htons(ETH_P_IP) || e->invflags & EBT_IPROTO)
  {
    return -EINVAL;
  }
  if (info->bitmask & ~EBT_IP_MASK || info->invflags & ~EBT_IP_MASK)
  {
    return -EINVAL;
  }
  if (info->bitmask & (EBT_IP_DPORT | EBT_IP_SPORT))
  {
    if (info->invflags & EBT_IP_PROTO)
    {
      return -EINVAL;
    }
    if (info->protocol != IPPROTO_TCP &&
        info->protocol != IPPROTO_UDP &&
        info->protocol != IPPROTO_UDPLITE &&
        info->protocol != IPPROTO_SCTP &&
        info->protocol != IPPROTO_DCCP)
    {
      return -EINVAL;
    }
  }
  if (info->bitmask & EBT_IP_DPORT && info->dport[0] > info->dport[1])
  {
    return -EINVAL;
  }
  if (info->bitmask & EBT_IP_SPORT && info->sport[0] > info->sport[1])
  {
    return -EINVAL;
  }
  return 0;
}

static struct xt_match EbModasMatch __read_mostly =
{
 .name        = "modas",
 .revision    = 0,
 .family      = NFPROTO_BRIDGE,
 .match       = EbModasMatchCB,
 .checkentry  = EbModasMatchCheckCB,
 .matchsize   = sizeof(struct EbtIpInfo_s),
 .me          = THIS_MODULE,
};

int EbModasInit(void)
{
  printk(KERN_INFO "EbModasInit\n");
  return xt_register_match(&EbModasMatch);
}

void EbModasExit(void)
{
  printk(KERN_INFO "EbModasExit\n");
  xt_unregister_match(&EbModasMatch);
}

MODULE_LICENSE("GPL");
MODULE_AUTHOR("Thomas Graf <tg@modas.de>");
